/*
 * Filter to convert row ASCII text to Structured PostScript
 * document.
 *
 * 10-01-92 10:09am
 *
 * Fixed by Vadim Maslov vadik@cs.umd.edu. Dec. 30, 1994.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifdef UNIX
#define SEEK_END	2
#endif

#define SHIFT_X 60

typedef unsigned char byte;
#include "fontbody.h"		/* character's bitmap definitions*/

/* translation table for ALTERNATIVE->KOI8 */
int koi8_to_alt[64] =
{0xEE,0xA0,0xA1,0xE6,0xA4,0xA5,0xE4,0xA3,
0xE5,0xA8,0xA9,0xAA,0xAB,0xAC,0xAD,0xAE,
0xAF,0xEF,0xE0,0xE1,0xE2,0xE3,0xA6,0xA2,
0xEC,0xEB,0xA7,0xE8,0xED,0xE9,0xE7,0xEA,
0x9E,0x80,0x81,0x96,0x84,0x85,0x94,0x83,
0x95,0x88,0x89,0x8A,0x8B,0x8C,0x8D,0x8E,
0x8F,0x9F,0x90,0x91,0x92,0x93,0x86,0x82,
0x9C,0x9B,0x87,0x98,0x9D,0x99,0x97,0x9A};

/* ============================================= Function Prototypes */
int get_param_string();		/* process command line */
int message();						/* */
int error_message();
int print_help();         /* print help message */
long filesize();          /* get file size */
int init_params();
int init_params0();
int print_all();
int read_font_data();
int insert_font();
int print_font_body();
int write_footer();
int write_char();
int write_header();
/* .................. character translation ........................ */
int translate1();         /* translate str from KOI-8 to ALT */
int charpos();            /* find position of ch in str */
/* .................. process documents ............................ */
int make_doc_header();    
int make_doc_footer();
int make_page_number();

int print_char();
int print_str();
int PRINT_str();
int FF();                 /* finish current page */
int check_and_split();    /* check ^L and split input string */


/* ============================================= Global Variables    */
char	PS_font_name[128], 
			PS_font_data[128];
byte	FONT_data[4096];
int 	FLAG_wrap=0;
int		FLAG_font_data=0;
int		FLAG_font_size=0;
int		FLAG_text_encod=0;
int		FLAG_split_line=0;
int		TEXT_point_size;
int		PAGE_length=0;
int		WRAP_pos;
double	LINE_spacing_f=1.0;
char    str_input[1024], str_lost[1024], str_to_print[1024];
int	page_length;			/* page length */
int	LF;								/* size of line-feed */
int	top_marging;			/* top marging */
int	bottom_marging;		/* bottom marging */
int	lines_per_page;		/* lines per page */
int	PAGE_number;      /* current page number */
int	hit_page;         /* 1 if any character were put to current page*/
int	IX, IY;           /* current coordinates */
int WRAPPED;          /* 1 if there are some characters from previouse line */
int	POP_page;
int	FLAG_EOF; 				/* EOF flag */
int	Gets(/* char *str */);


main(argc,argv)
int argc;
char *argv[];
{
     int ierror;
     int DO_work=1;


	init_params0(); /* initialize settings */
	/* check  input, if error exit */
	if( (ierror=get_param_string(argc, argv)) ) error_message(ierror);
	if( (ierror=init_params())!=0 ) error_message(ierror);
	if( (ierror=make_doc_header()) ) error_message(ierror);


	WRAPPED = 0;  POP_page = 0;
	/* Main cycle */
	while( DO_work ){
		if(DO_work){
		   if(WRAPPED==0){
		       memset(str_input,sizeof(str_input),0);
		       Gets(str_input); /* get string */
		       if(str_input[strlen(str_input)-1] == (char)0x0D) str_input[strlen(str_input)-1] = (char)0x00;
		       if(FLAG_EOF) DO_work = 0; /* EOF reached */
		   }
		   if(FLAG_text_encod) translate1(str_input); /* translate if KOI-8 */
		   check_and_split(str_to_print,str_input);
		   if(IY<bottom_marging) FF();	/* print page */
		   PRINT_str(IX,IY,str_to_print);
		   IY-=LF;	/* goto next line */
		   if(POP_page) FF();
		}
	} /* while */
        make_doc_footer();
	exit(0);
}
/* ................................................................. */
int    init_params0()
{
strcpy(PS_font_name,"PCnormal"); /* default font */
PS_font_data[0] = (char)0;      /* default data */
TEXT_point_size = 10;
PAGE_number = 1;
hit_page = 0;
FLAG_EOF = 0;
IX = SHIFT_X;
return(0);
}
/* ................................................................. */
int     get_param_string(p_c, p_a)
int p_c; char *p_a[];
{
int  _index=1, _exit_flag=0;
char _str[12];
int  _result, _NEXT;
double _resultf;
_NEXT = 0;

while ( !_exit_flag && (_index <= p_c-1)) {
	_NEXT = 0;

      if( strcmp(p_a[1],"/h")==0 )     return(332); /* short help */
      if( strcmp(p_a[1],"/help")==0 )  return(332); /* short help */
      if( strcmp(p_a[1],"-help")==0 )  return(332); /* short help */
      if( strcmp(p_a[1],"-h")==0 )     return(332); /* short help */

      if( strcmp(p_a[1],"/?")==0 )     return(333); /* short help */
      if( strcmp(p_a[1],"?")==0 )      return(333); /* short help */
      if( strcmp(p_a[1],"-?")==0 )     return(333); /* short help */

      if( strcmp(p_a[_index],"-f")==0 ){ /* next parameter font name */
         _index++;
         if(_index< p_c) strcpy(PS_font_name,p_a[_index]);
         else return(2);
	 if( PS_font_name[0] == '-' ) return(3);
	 _index++;
         _NEXT++;
	 if(_index >= p_c) return(0);
      }/* check         -f font_name     */

      if( strcmp(p_a[_index],"-Fd")==0 ){ /* next parameter font name */
         _index++;
	 if(_index< p_c) strcpy(PS_font_data,p_a[_index]);
         else return(2);
	 if( PS_font_name[0] == '-' ) return(4);
	 if(filesize(PS_font_data)!= 4096L)  return(5);
	 FLAG_font_data = 1;
	 _index++;
         _NEXT++;
	 if(_index >= p_c) return(0);
      }/* check         -Fd font_data    */

      if( strcmp(p_a[_index],"-ek")==0 ){
	 FLAG_text_encod = 1; /* do koi-8 encoding */
	 _index++;
         _NEXT++;
	 if(_index >= p_c) return(0);
      }/* check text encoding */

      if( strcmp(p_a[_index],"-s")==0 ){
         _index++; /* next argument number == point_size */
         if(_index< p_c) _result = atoi(p_a[_index]);
         else return(6); /* No point size is given */
         if((_result<0)||(_result>256))  return(7); /* invalid point size */
         TEXT_point_size = _result;
         _index++;
         _NEXT++;

	 if(_index >= p_c) return(0);
      }/* check point size */

      if( strcmp(p_a[_index],"-pl")==0 ){
         _index++; /* next argument number == page_length */
         if(_index< p_c) _result = atoi(p_a[_index]);
         else return(8); /* No page length is given */
         if(_result<0)  return(9); /* invalid page length */
         PAGE_length = _result;
         _index++;
         _NEXT++;
	 if(_index >= p_c) return(0);
      }/* page length */




      if( strcmp(p_a[_index],"-ls")==0 ){
         _index++; /* next argument number == line spacing */
         if(_index< p_c)
               _resultf = atof(p_a[_index]);
         else
              return(10); /* No line spacing is given */
	 if(_resultf<0.5)  return(11); /* invalid linr spacing */
         LINE_spacing_f = _resultf;
         _index++;
         _NEXT++;
	 if(_index >= p_c) return(0);
      }/* line spacing */

      if( strcmp(p_a[_index],"-X")==0 ){
	 _index++; /* next argument X shift */
         if(_index< p_c) _result = atoi(p_a[_index]);
	 else return(12); /* No page x shift is given */
	 if(_result<0)  return(13); /* invalid page length */
	 IX = _result;
         _index++;
         _NEXT++;
	 if(_index >= p_c) return(0);
      }/* shift X */

           if(_NEXT==0) _index++;
	}
   return(0);
}
/* ................................................................. */
int    print_help(help_mode)
int help_mode;
{
	switch (help_mode){
	  case 1: { /* short help */
       printf("TXT->PS filter v.01 (beta)\n");
       printf("USAGE: txt2ps [switches] < input_txt_file > output_ps_file\n");
       printf(" switches are:\t\t\tdefault\n");
       printf(" -f font_name\t\t\tPCnormal\n");
       printf(" -Fd font_data\n");
       printf(" -ek\tKOI encoding\t\talternative\n");
       printf(" -s  point_size\t\t\t10\n");
       printf(" -pl page_length\t\t70\n");
       printf(" -ls line_spacing\t\t1.0\n");
       printf(" -X horizontal shift\t\t%d\n",SHIFT_X);
       printf(" internal fonts are PCnormal and PCbold\n");
       printf("\ntype txt2ps /? for more help\n");
		   break;}
	  case 2: { /* long help  */
       printf("TXT->PS filter v.01 (beta)\n");
       printf(" THIS is BETA version of txt2ps filter, no more help available\n");
       printf(" Please e-mail to author zuk@elvis.sovusa.com or zuk@elvis.msk.su\n");
       printf(" all bugs and incompatibilities\n");
		   break;}
	}
	return(0);
}
/* ................................................................. */
int    message(error_message)
char *error_message;
{
	printf("ERROR. %s\n",error_message);
	return(0);
}
/* ................................................................. */
int    error_message(inumber)
int inumber;
{
	switch(inumber){
      case   1:  { message("number 1"); break;}
      case   2:  { message("No font name after -f"); break;}
      case   3:  { message("Invalid font name for -f"); break;}
      case   4:  { message("Invalid font name for -Fd"); break;}
      case   5:  { message("Invalid font data for -Fd"); break;}
      case   6:  { message("No point size for -s"); break;}
      case   7:  { message("Ivalid point size"); break;}
      case   8:  { message("No page length for -pl"); break;}
      case   9:  { message("Invalid page length"); break;}
      case  10:  { message("No line spacing for -ls"); break;}
      case  11:  { message("Invalid line spacing"); break;}
      case  12:  { message("No argument for X shift "); break;}
      case  13:  { message("Invalid X shift"); break;}
      case 332:  { print_help(1);       break;}
      case 333:  { print_help(2);       break;}

	}
	exit(1);
}
/* ................................................................. */
long    filesize(file_name)
char *file_name;
{
	int _fh;
	long _result=0L;
#ifndef UNIX
	if((_fh=open(file_name,O_RDONLY||O_BINARY))==-1) return(0L);
#else
	if((_fh=open(file_name,O_RDONLY))==-1) return(0L);
#endif
	if( (_result = lseek(_fh,0L,SEEK_END))==-1L ) _result = 0L;
	close(_fh);
	return(_result);
}
/* ................................................................. */
int     init_params()
{
	if(FLAG_font_data && (strstr(PS_font_name,"PCfixed")!=NULL) )
		      strcpy(PS_font_name,"UserCustomFont");

	/* calculae page layout settings */
	page_length = 792;
	/* set line feed to character height */
	LF = (int) (LINE_spacing_f*TEXT_point_size);
	top_marging = 2*LF;	/* set top marging to 2LF*/
	bottom_marging = 2*LF;
	lines_per_page = (page_length-
			  (top_marging+bottom_marging) ) / LF;

	if(PAGE_length&&(lines_per_page>PAGE_length)){
		bottom_marging = page_length - top_marging - PAGE_length*LF;
	}
	IY = 792 - top_marging - TEXT_point_size;
	WRAP_pos = (80*TEXT_point_size) / 10;
	return(0);
}
/* ................................................................. */
int     make_doc_header()
{
  printf("%%!PS-Adobe-2.0\n\n");
	printf("%%%%Creator: Witty user\n");
	printf("%%%%Title: User Document\n");
	printf("%%%%CreationDate: ????\n");
	printf("%%%%Pages: (atend)\n");
	printf("%%%%BoundingBox: 0 0 612 792\n");
	printf("%%%%EndComments\n");
	/* insert internal font or insert string to find... scale...set user custom font */
	insert_font(PS_font_name,PS_font_data);
	return(0);
}
/* ................................................................. */
int    print_all(){
       printf(" font_name %s\n",PS_font_name);
       if( *PS_font_data != 0) printf(" font data file %s\n",PS_font_data);
       if(FLAG_text_encod) printf(" KOI-8 encoding\n");
       printf(" %d point size\n",TEXT_point_size);
       if(PAGE_length) printf("  page length %d lines\n",PAGE_length);
       if( LINE_spacing_f!=1.0 ) printf("line_spacing %f \n",LINE_spacing_f);
       return(0);
}
/* ................................................................. */
int     read_font_data(font_name)
char *font_name;
{
	int _fh, _result;
#ifndef UNIX
	if((_fh=open(font_name,O_RDONLY||O_BINARY))==-1) return(4);
#else
	if((_fh=open(font_name,O_RDONLY))==-1) return(4);
#endif
	if((_result=read(_fh,FONT_data,4096))!=4096) return(5);
	close(_fh);
	return(0);
}
/* ................................................................. */
int     insert_font(font_name, file_name)
char *font_name;
char *file_name;
{
	if(FLAG_font_data){
	   if(read_font_data(file_name)) return(5); /* read data to buffer */
	   print_font_body(PS_font_name,FONT_data);
	}
	else{
	  if( strstr(font_name,"PCnormal") != NULL ) print_font_body(font_name,PC8x16_thik);
	  else
	     if( strstr(font_name,"PCbold") != NULL )   print_font_body(font_name,PC8x16_bold);
	     else
		printf("/%s findfont %d scalefont setfont\n",PS_font_name,TEXT_point_size);
	}
        printf("%%%%EndProlog\n");
	return(0);
}
/* ................................................................. */
int     print_font_body(font_name,font_buffer)
char *font_name;
byte *font_buffer;
{
	int _index;
	byte *ptr;

	write_header();		/* write font header */
	ptr =font_buffer;
	ptr +=16;
	for(_index=1;_index<256;_index++){
	   write_char(_index, ptr);
	   ptr+=16; /* go to next char def */
	}
	write_footer(font_name);	/* write font footer */
	/* define desired point size */
    printf("/%s findfont %d scalefont setfont\n\n",PS_font_name,TEXT_point_size);

}
/* ................................................................. */
int     write_footer(font_name)
char *font_name;
{
	printf("%%=====================================================\n");
	printf("end   %% of CharProcs\n\n");
	printf("/BuildChar\n{\n\t562 0\n\t0 0 1000 1000\n\tsetcachedevice\n");
	printf("\texch begin\n\t  Encoding exch get\n\t  CharProcs exch get\n");
	printf("\n\tend\n\texec\n}def\n\n");
	printf("end   %% of newfont\n");
	printf("/%s newfont definefont pop\n",font_name);
      return(0);
}
/* ................................................................. */
int   write_header()
{
      int result=0;
      int _ch_index;

printf("%% Bitmap font from 8x16 bitmap\n");
printf("%% Copyright (c). 1992, Yury Merezhuk.<zuk@elvis.msk.su>\n");
printf("/newfont 10 dict def\n");
printf("newfont begin\n\n");
printf("/FontType 3 def\n");
printf("/FontMatrix [.001 0 0 .001 0 0] def\n");
printf("/FontBBox [0 0 1000 1000] def\n\n");
printf("/Encoding 256 array def\n");
printf("0 1 255 {Encoding exch /.notdef put} for\n\n");

/* write 0 .. 32*/

for (_ch_index = 1; _ch_index<33; _ch_index++) {
    printf(
    "Encoding %3d /char_%02X put    %% ascii not printable,  = %d,\n",
    _ch_index,_ch_index,_ch_index);
}

/* write 33 .. 126*/
for (_ch_index = 33; _ch_index<127; _ch_index++) {
    printf(
    "Encoding %3d /char_%02X put    %% ascii '%c',  = %d,\n",
    _ch_index,_ch_index,_ch_index,_ch_index);
}

/* write 127 .. 255*/
for (_ch_index = 127; _ch_index<256; _ch_index++) {
    printf(
    "Encoding %3d /char_%02X put    %% ascii not printable,  = %d,\n",
    _ch_index,_ch_index,_ch_index);
}

printf("\n/CharProcs 256 dict def\n\n");
printf("CharProcs begin\n");
printf("%%=================================================\n");
printf("/.notdef {} def         %% for notdefined charactes\n");
/* .... set return(0) !!! ... */
      return(0);
}
/* ................................................................. */
int   write_char(char_index, char_def)
int char_index;
 byte *char_def;
{
      int read_count;

     printf("/char_%02X\n",char_index);
     if ((char_index<0xB0) || (char_index>0xDF))
        printf("{ 500 1000 scale\n8 16 true [8 0 0 -16 0 16]\n{<");
     else /* for pseudo-graphics symbols another CharBoundingBox */
        printf("{ 562 1000 scale\n8 16 true [8 0 0 -16 0 16]\n{<");

     for ( read_count=0;read_count<16;read_count++)
	   printf("%02X ", char_def[read_count]);
     printf(">} imagemask\n}def\n\n");

      return(0);
}
/* ................................................................. */
int     make_doc_footer()
{
	if(hit_page){
		printf("showpage\n");
		printf("%%%%EndPage: %d %d\n",PAGE_number,PAGE_number);
	}
        printf("%%%%Trailer\n");
        printf("%%%%DocumentFonts: %s\n",PS_font_name);
	printf("%%%%Pages: %d\n",PAGE_number);
	return(0);
}
/* ................................................................. */
int     make_page_number()
{
	printf("%%%%Page: %d %d\n",PAGE_number,PAGE_number);

	return(0);
}
/* ................................................................. */
int    translate1(str)
char *str;
{
        int _index;
        while ( *str != 0)
        {
           if ( (_index=(unsigned char)*str) >= 0xC0 )
		*str = (char)koi8_to_alt[ ((unsigned char)*str-0xC0) ];
           str++;
        }
}
/* ................................................................. */
int    print_char(ch) byte ch;
{
        if (ch=='\\' || ch>' ' && ch<0x2f) {
          printf("\\%03o", (byte) ch);
        } else {
          printf("%c",ch);
        }
	return(0);
}
/* ................................................................. */
int    print_str(str)
char *str;
{
        char _ch;
        while((_ch = *str++) != 0) {
          print_char(_ch);
        }
	return(0);
}
/* ................................................................. */
int    PRINT_str(ix, iy, str)
int ix;
int iy;
char *str;
{
	if(hit_page == 0){
	   hit_page = 1;
	   make_page_number();
	}
	if(*str != (char)0){
	  printf("%d %d moveto\n",ix,iy);
	  printf("(");
	  print_str(str);
	  printf(") show\n");
	}
	return(0);
}
/* ................................................................. */
int     charpos(str, ch)
char *str;
int ch;
{
	char *ch_ptr;
	if( (ch_ptr=strchr(str,ch)) == NULL) return(-1);
	else
	  return((int)(ch_ptr-str));

}
/* ................................................................. */
int    check_and_split(str1, str2)
char *str1;
char *str2;
{
	int _index;
	/* search for FF character */
	if( (_index=charpos(str2,12)) !=-1 ){
		strncpy(str1,str2,_index);
		str1[_index] = (char)0;
		/* remove first _index character from str2 */
		strcpy(str2,&(str2[_index+1]));
		WRAPPED = 1; POP_page = 1;
	}
	else{
	   if((_index=strlen(str2)) > WRAP_pos){
		strncpy(str1,str2,_index);
		str1[_index] = (char)0;
		/* remove first _index character from str2 */
		strcpy(str2,&(str2[_index]));
		WRAPPED = 1;
	   }
	   else{
		strcpy(str1,str2);  /* short string only copy */
		if(WRAPPED) WRAPPED = 0; /* no lost in buffer */
		str2[0]=(char)0;
	   }
	}
}
/* ................................................................. */
int    FF()
{
	IY = 792 - bottom_marging - TEXT_point_size;
	printf("showpage\n");
	printf("%%%%EndPage: %d %d\n",PAGE_number,PAGE_number);
	PAGE_number++;             /* switch to next page */
	POP_page = 0;
	hit_page = 0;
}
/* ................................................................. */
int    Gets(str)
char *str;
{
	int _index=0, _char;


	while( ((_char=getchar()) != 0x0A) && (_char != -1 ) ){
	   *str++ = (char)_char;
	}
	*str=(char)0x00; /* set terminated NULL */
	if(_char == -1) FLAG_EOF = 1;
	return(0);
}
